name = 'ssos'
version = '1.1.0'


GREETING = '''
       *       +      ssos pipeline %s
   +        x
        '             https://sso-pipeline.readthedocs.io
*   +     _.::'   |
        :_.'    - o - ssos --default : Copy default settings to CWD
|                 |   ssos --help    : Print detailed help
O -     *        '    ssos [-PARAMETER] [-VALUE] path/to/images
|
        _..:'         max.mahlke(at)cab.inta-csic.es
  +   (_.:

        ''' % version